/*-
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "vdef.h"
#include "vas.h"
#include "miniobj.h"

#include "ph.h"

/*
 * There is a non-zero probability that no perfect hash can be found,
 * infinitesimal for moderate to large sets, but not impossible for small
 * sets, if the tables are about as small as the set. Tests have never
 * shown hash generation failure when the tables have this minimum size.
 */
#define MINBITS (8)

struct ph {
	unsigned	magic;
#define PH_MAGIC	0x00cd8c1d
	int32_t		*inter;
	unsigned	*tbl;
	size_t		minlen;
	size_t		maxlen;
	unsigned	mask;
	unsigned	bits;
};

struct bucket {
	unsigned	*idx;
	int		n;
};

#define FNV64_OFFSET_BASIS (0xcbf29ce484222325)
#define FNV64_PRIME (0x00000100000001b3)

/* FNV-1a 64-bit with xor-folding */
static inline uint32_t
hash(uint32_t h, const char *s, size_t len, unsigned bits, unsigned mask)
{
	uint64_t h64 = h;

	if (h64 == 0)
		h64 = FNV64_OFFSET_BASIS;
	while (len--) {
		h64 ^= (uint64_t)*s++;
		h64 *= FNV64_PRIME;

	}
	return (((h64 >> bits) ^ h64) & mask);
}

/*
 * https://graphics.stanford.edu/~seander/bithacks.html#IntegerLogObvious
 */
static unsigned
lg(unsigned n)
{
	unsigned lg = 0;
	while (n >>= 1)
		lg++;
	return (lg);
}

/*
 * Sort buckets by length in *descending* order, which is the opposite of
 * the usual sense of qsort(). So:
 *
 * b1 > b2 -> return < 0
 * b1 < b2 -> return > 0;
 */
static int
len_desc_cmp(const void *b1, const void *b2)
{
	AN(b1);
	AN(b2);

	return ( ((struct bucket *)b2)->n - ((struct bucket *)b1)->n);
}

struct ph *
PH_Generate(char * const * const strings, unsigned n)
{
	size_t min = SIZE_MAX, max = 0, len;
	int *found = NULL;
	struct bucket *buckets = NULL;
	int32_t *inter = NULL;
	unsigned bits, sz, mask, i, *tbl = NULL;
	struct ph *ph = NULL;

	AN(strings);
	assert(n > 0 && n < 1 << 30);

	bits = lg(n);
	if (n != (unsigned)(1 << bits))
		bits++;
	if (bits < MINBITS)
		bits = MINBITS;
	sz = 1 << bits;
	mask = sz - 1;

	errno = 0;
	buckets = calloc(sz, sizeof(*buckets));
	if (buckets == NULL)
		return (NULL);
	tbl = malloc(sz * sizeof(*tbl));
	if (tbl == NULL)
		goto exit;
	memset(tbl, 0xff, sz * sizeof(*tbl));
	inter = calloc(sz, sizeof(*inter));
	if (inter == NULL)
		goto exit;
	found = malloc(sz * sizeof(*found));
	if (found == NULL)
		goto exit;

	for (i = 0; i < n; i++) {
		uint32_t h;

		len = strlen(strings[i]);
		h = hash(0, strings[i], len, bits, mask);
		assert(h < sz);

		errno = 0;
		buckets[h].idx = realloc(buckets[h].idx,
					 (buckets[h].n + 1)
					 * sizeof(*buckets[h].idx));
		if (buckets[h].idx == NULL)
			goto exit;
		buckets[h].idx[buckets[h].n] = i;
		buckets[h].n++;

		if (len < min)
			min = len;
		if (len > max)
			max = len;
	}

	qsort(buckets, sz, sizeof(*buckets), len_desc_cmp);

	for (i = 0; i < sz && buckets[i].n > 1; i++) {
		struct bucket *bucket = &buckets[i];
		uint32_t m = 1;

		AN(bucket->idx);
		memset(found, 0, sz * sizeof(*found));

		for (int j = 0; j < bucket->n && m < UINT32_MAX; j++) {
			uint32_t h;

			len = strlen(strings[bucket->idx[j]]);
			h = hash(m, strings[bucket->idx[j]], len, bits, mask);
			assert(h < sz);
			if (tbl[h] != UINT_MAX || found[h] != 0) {
				m++;
				j = -1;
				memset(found, 0, sz * sizeof(*found));
				continue;
			}
			found[h] = 1;
		}
		if (m == UINT32_MAX) {
			errno = ERANGE;
			goto exit;
		}

		len = strlen(strings[bucket->idx[0]]);
		inter[hash(0, strings[bucket->idx[0]], len, bits, mask)] = m;
		for (int j = 0; j < bucket->n; j++) {
			len = strlen(strings[bucket->idx[j]]);
			tbl[hash(m, strings[bucket->idx[j]], len, bits, mask)]
				= bucket->idx[j];
		}
	}

	/* Reuse found[] for a list of free indices in tbl. */
	memset(found, 0, sz * sizeof(*found));
	for (unsigned j = 0, n = 0; j < sz; j++)
		if (tbl[j] == UINT_MAX)
			found[n++] = j;

	/*
	 * Continuing with the value of i from the for loop above, now at
	 * buckets[i].n <= 1
	 */
	for (int n = 0; i < sz && buckets[i].n > 0; n++, i++) {
		struct bucket *bucket = &buckets[i];

		len = strlen(strings[bucket->idx[0]]);
		tbl[found[n]] = bucket->idx[0];
		inter[hash(0, strings[bucket->idx[0]], len, bits, mask)]
		      = -found[n] - 1;
	}

	errno = 0;
	ALLOC_OBJ(ph, PH_MAGIC);
	if (ph == NULL)
		goto exit;
	ph->inter = inter;
	ph->tbl = tbl;
	ph->mask = mask;
	ph->bits = bits;
	ph->minlen = min;
	ph->maxlen = max;

 exit:
	AN(buckets);
	for (i = 0; i < sz; i++)
		if (buckets[i].idx != NULL)
			free(buckets[i].idx);
	free(buckets);
	if (found != NULL)
		free(found);
	if (ph == NULL) {
		if (inter != NULL)
			free(inter);
		if (tbl != NULL)
			free(tbl);
	}

	return (ph);
}

unsigned
PH_Lookup(const struct ph * const restrict ph,
	  char * const restrict * const restrict strings,
	  const char * const restrict subject)
{
	size_t len;
	int32_t h;
	unsigned idx;

	CHECK_OBJ_ORNULL(ph, PH_MAGIC);
	AN(strings);
	AN(subject);

	if (ph == NULL)
		return (UINT_MAX);

	len = strlen(subject);
	if (len < ph->minlen)
		return (UINT_MAX);
	if (len > ph->maxlen)
		return (UINT_MAX);

	h = ph->inter[hash(0, subject, len, ph->bits, ph->mask)];
	if (h < 0)
		idx = ph->tbl[-h - 1];
	else
		idx = ph->tbl[hash(h, subject, len, ph->bits, ph->mask)];

	if (idx == UINT_MAX || strcmp(subject, strings[idx]) != 0)
		return (UINT_MAX);
	return (idx);
}

void
PH_Free(struct ph *ph)
{
	if (ph == NULL)
		return;
	free(ph->inter);
	free(ph->tbl);
	FREE_OBJ(ph);
}
